@@ -25,7 +25,8 @@
 #include <config.h>
 #endif
 #include "Xrenderint.h"
- 
+#include <limits.h>
+
 XFilters *
 XRenderQueryFilters (Display *dpy, Drawable drawable)
 {
@@ -37,8 +38,8 @@ XRenderQueryFilters (Display *dpy, Drawable drawable)
     char			*name;
     char			len;
     int				i;
-    long			nbytes, nbytesAlias, nbytesName;
-    
+    unsigned long		nbytes, nbytesAlias, nbytesName, reply_left;
+
     if (!RenderHasExtension (info))
 	return NULL;
 
@@ -48,7 +49,7 @@ XRenderQueryFilters (Display *dpy, Drawable drawable)
     xri = info->info;
     if (xri->minor_version < 6)
 	return NULL;
-    
+
     LockDisplay (dpy);
     GetReq (RenderQueryFilters, req);
     req->reqType = info->codes->major_opcode;
@@ -60,26 +61,36 @@ XRenderQueryFilters (Display *dpy, Drawable drawable)
 	SyncHandle ();
 	return NULL;
     }
+
     /*
-     * Compute total number of bytes for filter names
+     * Limit each component of combined size to 1/4 the max, which is far
+     * more than they should ever possibly need.
      */
-    nbytes = (long)rep.length << 2;
-    nbytesAlias = rep.numAliases * 2;
-    if (rep.numAliases & 1)
-	nbytesAlias += 2;
-    nbytesName = nbytes - nbytesAlias;
-    
-    /*
-     * Allocate one giant block for the whole data structure
-     */
-    filters = Xmalloc (sizeof (XFilters) +
-		       rep.numFilters * sizeof (char *) +
-		       rep.numAliases * sizeof (short) +
-		       nbytesName);
+    if ((rep.length < (INT_MAX >> 2)) &&
+	(rep.numFilters < ((INT_MAX / 4) / sizeof (char *))) &&
+	(rep.numAliases < ((INT_MAX / 4) / sizeof (short)))) {
+	/*
+	 * Compute total number of bytes for filter names
+	 */
+	nbytes = (unsigned long)rep.length << 2;
+	nbytesAlias = rep.numAliases * 2;
+	if (rep.numAliases & 1)
+	    nbytesAlias += 2;
+	nbytesName = nbytes - nbytesAlias;
+
+	/*
+	 * Allocate one giant block for the whole data structure
+	 */
+	filters = Xmalloc (sizeof (XFilters) +
+			   (rep.numFilters * sizeof (char *)) +
+			   (rep.numAliases * sizeof (short)) +
+			   nbytesName);
+    } else
+	filters = NULL;
 
     if (!filters)
     {
-	_XEatData (dpy, (unsigned long) rep.length << 2);
+	_XEatDataWords(dpy, rep.length);
 	UnlockDisplay (dpy);
 	SyncHandle ();
 	return NULL;
@@ -92,7 +103,7 @@ XRenderQueryFilters (Display *dpy, Drawable drawable)
      *	numAliases  short alias values
      *	nbytesName  char strings
      */
-    
+
     filters->nfilter = rep.numFilters;
     filters->nalias = rep.numAliases;
     filters->filter = (char **) (filters + 1);
@@ -103,6 +114,7 @@ XRenderQueryFilters (Display *dpy, Drawable drawable)
      * Read the filter aliases
      */
     _XRead16Pad (dpy, filters->alias, 2 * rep.numAliases);
+    reply_left = 8 + rep.length - 2 * rep.numAliases;;
 
     /*
      * Read the filter names
@@ -111,9 +123,19 @@ XRenderQueryFilters (Display *dpy, Drawable drawable)
     {
 	int	l;
 	_XRead (dpy, &len, 1);
+	reply_left--;
 	l = len & 0xff;
+	if ((unsigned long)l + 1 > nbytesName) {
+            _XEatDataWords(dpy, reply_left);
+	    Xfree(filters);
+	    UnlockDisplay (dpy);
+	    SyncHandle ();
+	    return NULL;
+	}
+	nbytesName -= l + 1;
 	filters->filter[i] = name;
 	_XRead (dpy, name, l);
+        reply_left -= l;
 	name[l] = '\0';
 	name += l + 1;
     }
@@ -121,8 +143,9 @@ XRenderQueryFilters (Display *dpy, Drawable drawable)
 
     if (i & 3)
 	_XEatData (dpy, 4 - (i & 3));
-    
+
     UnlockDisplay (dpy);
+    SyncHandle ();
     return filters;
 }
 
